#!BPY

# texture2vertexcol.py 1.0
# Copyright (c) 2008 Christian Walther <cwalther@gmx.ch>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Name: 'Texture to Vertex Colors'
Blender: 246
Group: 'Mesh'
Tooltip: 'Approximate a texture image with vertex colors'
"""

import math
import numpy
import bpy
import Blender

def run():
	obj = bpy.data.scenes.active.objects.active
	if obj.type != "Mesh":
		Blender.Draw.PupMenu("Error%t|Please select a mesh.")
		return
		
	img = bpy.data.images.active
	if img == None:
		Blender.Draw.PupMenu("Error%t|Please select an image in the UV/image editor.")
		return
	w, h = img.getMaxXY()
	
	mesh = obj.getData(mesh=True)
# It would be nice to use the modified mesh to be able to get sharp edges, but we'd somehow have to get the vertex colors back onto the original mesh
#	mesh = Blender.Mesh.New()
#	mesh.getFromObject(obj)
	
	# add a vertex color layer if there isn't already one
	mesh.vertexColors = True
	
	# set up storage for collecting colors from each adjacent triangle for each vertex
	colors = [[] for i in range(len(mesh.verts))]
	
	for face in mesh.faces:
		# a quad must be split into 2 triangles (in the same way as PyPRP does it while exporting)
		if len(face) == 4:
			triangles = ((0, 1, 2), (0, 2, 3))
		else:
			triangles = ((0, 1, 2),)
		for tri in triangles:
			# convert normalized uv coordinates to pixel coordinates
			uv0 = face.uv[tri[0]]
			uv1 = face.uv[tri[1]]
			uv2 = face.uv[tri[2]]
			x0 = uv0.x*w
			y0 = uv0.y*h
			x1 = uv1.x*w
			y1 = uv1.y*h
			x2 = uv2.x*w
			y2 = uv2.y*h
			# precompute the matrix for conversion to barycentric coordinates
			xyToBary = numpy.linalg.inv(numpy.array([[x0-x2, x1-x2], [y0-y2, y1-y2]]))
			# find the bounding box of the triangle
			xmin = int(math.floor(min(x0, x1, x2)))
			xmax = int(math.ceil(max(x0, x1, x2)))
			ymin = int(math.floor(min(y0, y1, y2)))
			ymax = int(math.ceil(max(y0, y1, y2)))
			# allocate the matrices we're going to fill in, with enough space for the whole bounding box (I have a feeling that half of that would be sufficient, but I'm too lazy to prove that right now)
			pixelbs = numpy.empty(((xmax-xmin)*(ymax-ymin), 3)) # barycentric coordinates
			pixelvs = numpy.empty(((xmax-xmin)*(ymax-ymin), 3)) # r, g, b
			# iterate over all pixels in the bounding box
			count = 0
			for y in range(ymin, ymax):
				for x in range(xmin, xmax):
					# compute the barycentric coordinates of the pixel center
					b0, b1 = numpy.dot(xyToBary, numpy.array((x + 0.5 - x2, y + 0.5 - y2)))
					b2 = 1 - b0 - b1
					if b0 >= 0 and b1 >= 0 and b2 >= 0: # else it's outside of the triangle
						pixelbs[count, :] = (b0, b1, b2)
						pixelvs[count, :] = img.getPixelI(x, y)[:3]
						count += 1
			if count != 0: # ignore triangles that are so small that they have no pixels
				# truncate the matrices to their actually filled length
				pixelbs = numpy.vsplit(pixelbs, (count,))[0]
				pixelvs = numpy.vsplit(pixelvs, (count,))[0]
				# compute the linear regression and store the result
				res = numpy.linalg.lstsq(pixelbs, pixelvs)[0]
				colors[face.verts[tri[0]].index].append(res[0, :])
				colors[face.verts[tri[1]].index].append(res[1, :])
				colors[face.verts[tri[2]].index].append(res[2, :])
				
# applying the colors directly to this face, without averaging with adjacent faces, produces pretty discontinuous results (left here for debugging purposes)
#				for i in range(3):
#					r = res[i, :].round().astype(int)
#					r[r < 0] = 0
#					r[r > 255] = 255
#					c = face.col[i]
#					c.r, c.g, c.b = r
			
	# average the collected colors per vertex
	for i, l in enumerate(colors):
		if len(l) == 0:
			colors[i] = numpy.array((255, 255, 255))
		else:
			c = numpy.array(l).mean(0).round().astype(int)
			c[c < 0] = 0
			c[c > 255] = 255
			colors[i] = c
			
	# apply the averaged color to each vertex
	for face in mesh.faces:
		for i in range(len(face)):
			c = face.col[i]
			c.r, c.g, c.b = colors[face.verts[i].index]
			c.a = 255

	# not sure if this is necessary or sufficient to make the result visible
	mesh.update()


run()
